/**
 * Kode-fil for rute-klassen
 *
 * @file rute.cpp
 * @author Sondre Sand & Frederik Simonsen
 */
#include <algorithm> //all_of
#include <iostream>
#include <iomanip>
#include <list>
#include <fstream>
#include "const.h"
#include "rute.h"
#include "stoppested.h"
#include "stoppesteder.h"
#include "LesData3.h"
using namespace std;

extern Stoppesteder gStoppestederBase;

//********STOPP FUNKSJONER*******************


/**
 *  Construct a new Stopp:: Stopp object
 *
 * @param stoppNr
 * @param antMin
 */
Stopp::Stopp(const int stoppNr, const int antMin) {
    nr = stoppNr; minutter = antMin;
}

/**
 * Henter en stopps unike nr
 *
 * @return Et stoppesteds unike nr
 */
int Stopp::hentNr() {
    return nr;
}

/**
 * @brief henter minutter fra forrige stopp
 *
 * @return int
 */
int Stopp::hentMin() {
    return minutter;
}

/**
 * skrivTilFil(...)
 *
 * @param ut
 */
void Stopp::skrivTilFil(ofstream & ut) {
    ut << nr << ' ' << minutter << '\n';
}

//********RUTE FUNKSJONER*******************


/**
 * Rute()
 */
Rute::Rute() {
    // Gjør ingenting, men må være her så man kan opprette buss/bane
}

/**
 * Opretter en ny rute fra fil
 *
 * @param inn filen det leses fra
 */
Rute::Rute(ifstream & inn) {
    int antStopp,
        stoppNr,
        antMin;

    inn >> antStopp; inn.ignore();
    for (int i = 0; i < antStopp; i++) {
        inn >> stoppNr >> antMin; inn.ignore();
        stoppene.push_back(new Stopp(stoppNr, antMin));
    }
}

/**
 * -sjekker om lista er mer en 1 entry.
 *
 * @return true
 * @return false
 */
bool Rute::erListeGyldig(){
    if(stoppene.size() > 1){
        return true;
    }
    return false;
}

/**
 * @brief
 *
 * @param nvn
 * @return true
 * @return false
 */
bool Rute::erTall(std::string nvn) {
    return all_of(nvn.begin(), nvn.end(), ::isdigit);
}

/**
 * Sjekker om stopp er lagt til i ruta allerede
 *
 * @param nr
 * @return true
 * @return false
 * @see Stopp::hentNr()
 */
bool Rute::finnesStopp(const int indeks) {
    for (auto it = stoppene.begin(); it != stoppene.end(); it++) {
        if ((*it)->hentNr() == indeks)
            return true;
    }
    return false;
}

/**
 * Funksjon for å sjekke om et klokkeslett er gyldig (00:00-23:59)
 *
 * @param time
 * @param minutt
 * @return true
 * @return false
 */
bool Rute::gyldigTid(const int time, const int minutt) {
    if ((time >= 0 && time <= 23) && (minutt >= 0 && minutt <= 59)) {
        return true;
    } else
        return false;
}

int Rute::finnDifferanse(const int nr, const Retning retning) {
    int totMin = 0;
    if (retning == Fram) {
        for (auto it = stoppene.begin(); it != stoppene.end(); it++) {
            totMin += (*it)->hentMin();
            if ((*it)->hentNr() == nr)
                return totMin;
        }
    } else if (retning == Tilbake) {
        for (auto it = stoppene.rbegin(); it != stoppene.rend(); it++) {
            if ((*it)->hentNr() == nr)
                return totMin;
            totMin += (*it)->hentMin();
        }
    }
    return -1; // Skal ikke skje
}

string Rute::hentNavn(const Retning retning) {
    string startSted;
    if (retning == Fram) {
    startSted = gStoppestederBase.hentNavnVhaIndeks(stoppene.front()->hentNr());
    } else if (retning == Tilbake)
        startSted = gStoppestederBase.hentNavnVhaIndeks(stoppene.back()->hentNr());

   return startSted;
}

/**
 * 
 * 
 * @param diff - differanse mellom startsted og aktuelt stoppested i minutter
 * @param start - faktisk startstedsnavn på ruten
 * @param stSted - faktisk navn på aktuellt stoppested
 * @see Rute::gyldigTid(...)
 */
void Rute::ruteTabell(const int diff, const string start, const string stSted) {
    int startT, // starttime
        startTotal, // regner ut total starttid i minutter
        sluttT, // slutt-time
        sluttTotal, // regner ut total slutttid i minutter
        startM, // startminutter
        sluttM, // slutt-minutter
        tidMellom, // Tid mellom hver avgang i minutter
        tPrint, // hjelpevariabel for å sørge for korrekt utskrift av time
        antGanger, // omregning for å legge inn riktig antall avgangstider
        startTid, // omregning for å legge inn starttid
        n, // hjelpevariabel for linjeskift ved ny time
        timer, // hjelpevariabel for å skrive ut time
        minutter, // hjelpevariabel for å skrive ut minutter
        avgangsTid, // omregning for å legge inn riktig avgangstid
        mPrint; // hjelpevariabel for å sørge for korrekt utskrift av minutter
    vector <int> avgangstider; // Hjelpevektor for å ta vare på avgangstider

    cout << "\nRuteavganger fra startstedet (" << start << ") - "
         << "avslutt med 0 0):\n\n";

    cout << "Fra kl. (00:01 - 23:59): ";
    cin >> startT >> startM; cin.ignore(); // Leser starttid


    while (!gyldigTid(startT, startM)) { // Sørger for at starttid er gyldig
        cout << "\nUlovlig klokkeslett. Prøv igjen (tt mm): ";
        cin >> startT >> startM; cin.ignore();
    }
    startTid = (startT*60)+startM; // Regner ut første avgangstid
    n = startT; // Setter indikator til å være første starttime


    while (startT != 0 || startM != 0) { // Så lenge bruker ikke taster 0 0
        avgangstider.push_back(startTid+diff); // Legger start tid i vector
        // Leser tid mellom avganger
        tidMellom = lesInt("Tid mellom avganger", MINTID, MAXTID);
        startTotal = (startT*60)+startM+tidMellom; // Omregner for sjekk
        tPrint = startTotal/60; // Regner ut gyldig utskrift:
        mPrint = startTotal%60;

        cout << "Til kl. (" << ((tPrint < 10) ? "0" : "") << tPrint
             << ':' << ((mPrint < 10) ? "0" : "") << mPrint
             << " - 23:59): ";
        // Leser slutt tid
        cin >> sluttT >> sluttM; cin.ignore();
        sluttTotal = (sluttT*60)+sluttM; // Omregner for sjekk
        // Sørger for lovlig slutt-tid
        while (!gyldigTid(sluttT, sluttM) || sluttTotal < startTotal) {
            cout << "\nUlovlig klokkeslett. Prøv igjen (tt mm): ";
            cin >> sluttT >> sluttM; cin.ignore();
            sluttTotal = (sluttT*60)+sluttM; // Omregner for sjekk
        }

        antGanger = ((sluttT-startT)*60)/tidMellom; // Omregning for loop
        avgangsTid = startTid+diff; // Setter på differanse en gang

        for (int i = 0; i < antGanger; i++) { // Legger til riktig antall avgangstider:
            avgangsTid = avgangsTid + tidMellom; // Omregning
            avgangstider.push_back(avgangsTid); // Legges i vektor
        }

        tPrint = (sluttTotal+1)/60; // Regner ut gyldig utskrift:
        mPrint = (sluttTotal+1)%60;


        cout << "\n\nFra kl. (" << ((tPrint < 10) ? "0" : "") <<  tPrint
             << ':' << ((mPrint < 10) ? "0" : "") << mPrint
             << " - 23:59): ";
        cin >> startT >> startM; cin.ignore(); // Leser ny starttid
        startTotal = (startT*60)+startM; // Omregning
        // Sørger for ny lovlig starttid:
        while ((!gyldigTid(startT, startM) || startTotal <= sluttTotal)
                && (startT != 0 || startM != 0)) { // Sørger for at starttid er gyldig
            cout << "\nUlovlig klokkeslett. Prøv igjen (tt mm): ";
            cin >> startT >> startM; cin.ignore();
            startTotal = (startT*60)+startM; // Omregning
        }
        startTid = (startT*60)+startM; // Omregning
    }
    cout << "\n\nRutetabell for stoppested: " << stSted << "\n\n"
         << ((n < 10) ? "0" : "") << n << ':';

    for (int i = 0; i < avgangstider.size(); i++) { // For hver avgangstid:
        timer = avgangstider[i]/60; // Regner ut time
        minutter = avgangstider[i]%60; // Regner ut minutter
        if (timer == n) { // Skriver ut minutter hvis time er lik indikator
            cout << ' ' << ((minutter < 10) ? "0" : "") << minutter;
        } else { // Skriver ut time hvis timer er mer enn indikator
            cout << '\n' << ((timer < 10) ? "0" : "")
                 << timer << ':' << ' ' << ((minutter < 10) ? "0" : "")
                 << minutter;
            n = timer; // Oppdaterer indikator
        }
    }
}

/**
 * skrivRetninger(...) brukes med R T
 *
 */
void Rute::skrivRetninger() const {
    // Hjelpevariabler for å hente stasjonsnavnene knyttet til ruten:
    int startIndeks = stoppene.front()->hentNr(),
        stoppIndeks = stoppene.back()->hentNr();
    string startStasjon = gStoppestederBase.hentNavnVhaIndeks(startIndeks),
           endeStasjon = gStoppestederBase.hentNavnVhaIndeks(stoppIndeks);

    // Skriver ut start- og endestasjon for ruten i begge retninger:
    cout << "\nFram:    Retningen: " << startStasjon << " - "
         << endeStasjon << "\nTilbake: Retningen: "
         << endeStasjon << " - " << startStasjon << '\n';
}

/**
 * skriver rute forlengs eller baklengs
 *
 * @param retning forlengs eller baklengs
 * @see Stoppesteder::hentNavnVhaIndeks(...)
 */
void Rute::skrivRute(const Retning retning) {
    int minF = 0,
        indeks = 0,
        totMin = 0;
    string navn;

    if (retning == Fram) {
        for (auto it = stoppene.begin(); it != stoppene.end(); it++) {
            indeks = (*it)->nr;
            minF = (*it)->minutter;
            navn = gStoppestederBase.hentNavnVhaIndeks(indeks);
            totMin += minF;
            cout << setw(25) << navn << " - " << setw(2) << minF
                 << " min fra forrige  (" << setw(3) << totMin
                 << " min fra rutestart)\n";
        }
    }

    if (retning == Tilbake) {
        for (auto it = stoppene.rbegin(); it != stoppene.rend(); it++) {
            indeks = (*it)->nr;
            navn = gStoppestederBase.hentNavnVhaIndeks(indeks);
            cout << setw(25) << navn << " - " << setw(2) << minF
                    << " min fra forrige  (" << setw(3) << totMin
                    << " min fra rutestart)\n";
            minF = (*it)->minutter;
            totMin += minF;
        }
    }
}

/**
 * @brief - sletter siste entry i stoppene.
 *
 */
void Rute::slettData(){
    stoppene.clear();
}

/**
 * Leser og oppretter stopp for en ny rute
 *
 * @see Stoppesteder::byttBokstaver(...)
 * @see Stoppesteder::finnesIndex(...)
 * @see Stoppesteder::hentNavnVhaIndex(...)
 * @see Stoppesteder::finnEntydig(...)
 * @see Stoppesteder::finnDuplikat(...)
 * @see Rute::finnesStopp(...)
 * @see Stoppested::finnesNabo(...)
 * @see Stoppested::hentNaboIndeks(...)
 * @see Stoppested::hentNaboTid(...)
 * @see Stoppested::settNaboIndeks(...)
 * @see Stoppested::settNaboTid(...)
 */
void Rute::lesData() {
    string navn,
           fNavn,
           nNavn;
    int stoppNr = 0,
        indeks = 0,
        naboIndeks = 0,
        tidTilF = 0;
    Stopp* nyStopp = nullptr;
    Stoppested* fStopp = nullptr;
    Stoppested* nStopp = nullptr;

    cout << "\nStoppested (entydig navn / tall / ENTER for å avslutte): ";
    getline(cin, navn);
    while (navn.size() > 0) {  //Kjør så lenge ikke blank/enter er trykket.
        navn = gStoppestederBase.byttBokstaver(navn); // Fjerner æøåÆØÅ

        if (erTall(navn)) {  // SJekker om navn kun består av tall
            stoppNr = stoi(navn); //Hvis kun tall, gjør om til int
            if(gStoppestederBase.finnesIndeks(stoppNr-1)){ //@see finnesIndeks, funker dette?
            navn = gStoppestederBase.hentNavnVhaIndeks(stoppNr); // Hent navn
            }
    }

        fNavn = gStoppestederBase.finnEntydig(navn); // Sjekker for entydighet

        if (fNavn.size() > 0) { // Hvis entydig navn:
            cout << '\n' << fNavn << '\n';
            indeks = gStoppestederBase.hentIndeksVhaNavn(fNavn);
            if (!finnesStopp(indeks+1)) { // Hvis stopp ikke finnes fra før på ruten
            if (stoppene.size() > 0) { // Hvis ikke første stopp på ruten
                nStopp = gStoppestederBase.finnDuplikat(fNavn);
                if (nStopp->finnesNabo(naboIndeks+1)) { // Hvis nabo finnes fra før
                    naboIndeks = nStopp->hentNaboIndeks(naboIndeks+1);
                    tidTilF = nStopp->hentNaboTid(naboIndeks);
                    nyStopp = new Stopp(indeks+1, tidTilF);
                    stoppene.push_back(nyStopp);
                    naboIndeks = indeks;
                    cout << "\nTiden mellom stoppestedene er allerede registrert som "
                         << tidTilF << " minutter.\n";


                } else { // Hvis nabo ikke finnes fra før:
                    tidTilF = lesInt("Tid fra forrige stopp", 1, 10); // Leser tid til forrige nabo
                    nStopp = gStoppestederBase.finnDuplikat(fNavn); // Finner egen peker
                    fStopp->settNaboIndeks(indeks); // Oppdaterer forrige stopps nabovektor
                    fStopp->settNaboTid(tidTilF); // Oppdaterer forrige stopps tid til nabo
                    nStopp->settNaboIndeks(naboIndeks); // Oppdaterer nåværende stopps nabovektor
                    nStopp->settNaboTid(tidTilF); // Oppdaterer naværende stopps tid til nabo
                    nyStopp = new Stopp(indeks+1, tidTilF); // Oppretter ny stopp på ruten
                    stoppene.push_back(nyStopp); // Stoppen legges til ruten
                    naboIndeks = indeks;
                }

            } else { // Hvis første stopp på ruten
                nStopp = gStoppestederBase.finnDuplikat(fNavn); // Peker til riktig
                nyStopp = new Stopp(indeks+1, 0); // Setter stopp sitt nr
                // til å korrespondere med indeks i vector, og tid til
                // forrige nabo å være 0, da første stopp på ruten
                stoppene.push_back(nyStopp); // Legges til listen
                naboIndeks = indeks;
            }
            nNavn = fNavn; // Setter nåværende nabo til å være forrige navn
            fStopp = nStopp; // Setter forrige stopp til å bli nåværende stopp
            } else
                cout << "\nStoppet er allerede registrert på ruten!\n";
        } else { // Hvis ikke entydig:
            cout << "\nIkke funnet (den entydige) stoppestedet!\n";
        }

        // Forsøker å lese evt. ny eller korrekt stopp:
        cout << "\nStoppested (entydig navn / tall / ENTER for å avslutte): ";
        getline(cin, navn);
    }
    cout << "\nOk, du angrer og ønsker ikke å legge til ny rute.\n";
}

/**
 * skrivBeskrivelse()
 * skriver forlengs og baklengs
 *
 * @see Rute::skrivRute(...)
 */
void Rute::skrivBeskrivelse() {
    cout << "Ruten:\n";
    skrivRute(Fram);
    cout << "\n\n";
    skrivRute(Tilbake);
}

/**
 * Skriver ut en rute
 * Første-siste stopp
 *
 * @see Stoppesteder::hentNavnVhaIndeks(...)
 */
void Rute::skrivData() const {
    int startIndeks,
        stoppIndeks;
    string startNavn,
           stoppNavn;

    startIndeks = stoppene.front()->hentNr(); // Henter startsteds unike nr
    stoppIndeks = stoppene.back()->hentNr(); // Henter stopsteds unike nr
    startNavn = gStoppestederBase.hentNavnVhaIndeks(startIndeks);
    stoppNavn = gStoppestederBase.hentNavnVhaIndeks(stoppIndeks);

    cout << startNavn << " - " << stoppNavn;
}

/**
 * @brief skrivKort() (virtuell)
 *
 */
void Rute::skrivKort() const {

}

/**
 * skriver rute til fil
 *
 * @param ut fila det skrives til
 * @see Stopp::skrivTilFil(...)
 */
void Rute::skrivTilFil(ofstream & ut) {
    ut << stoppene.size() << '\n';
    for (auto it = stoppene.begin(); it != stoppene.end(); it++)
        (*it)->skrivTilFil(ut);
}













